from __future__ import division

from scipy.stats import norm, gamma as gammaDist, beta as betaDist
from scipy.special import gammaln
from random import random
from numpy import zeros, array
import pylab
from cookbook.frange import frange
from math import sqrt, log, exp
import cPickle as pickle
from copy import copy

import sys

from model import normal, bernoulli, plotData

OPTIONS = [
#    "bad-gibbs", #sample z from the wrong distr
#    "greedy-gibbs", #choose z greedily
#    "guided-mh", #asymmetric proposal examining cluster means
#    "reverse-mhr", #reverse mh ratio to choose bad alternatives
#    "greedy-mh", #accept mh proposals greedily,
#    "single-cluster", #put everything in a single cluster to start
#    "true-params", #fix mu, sigma at true values
#    "far-away", #start cluster centers far from data
#    "bad-init", #try to find a local-max configuration
#    "sigmaq-small", #set sigma_q small
#    "sigmaq-large", #set sigma_q large
    ]

def sampleZ(xs, ms, vs, zs, u=1, v=1):
    #collapsed Gibbs for z
    n = len(xs) #sufficient stats of the zs
    n1 = sum(zs)
    n0 = n - n1
    ni = [n0 + u, n1 + v]

    ll = 0

    for ii,xi in enumerate(xs):
        ni[zs[ii]] -= 1 #remove z_i to find sufficient stats of z_{-i}
        ps = []

        for cluster in [0,1]:
            #compute p(z, x | mu, sigma, z_{-i})
            pJoint = 1.0
            for coord in [0,1]:
                #p(x | z=cluster...)
                pJoint *= norm.pdf(xi[coord], ms[cluster][coord],
                                   vs[cluster][coord][coord])
            #p(z=cluster | z_{-i})
            pJoint *= ni[cluster] / sum(ni)
            ps.append(pJoint)
        #marginalize to obtain p(z=cluster | x, ...)
        ps = [pp / sum(ps) for pp in ps]
        if "bad-gibbs" in OPTIONS:
            #if this block is active, the Gibbs sampler chooses an
            #alternative with 1 - the correct probability
            zNew = bernoulli(ps[0])
        elif "greedy-gibbs" in OPTIONS:
            #greedy gibbs
            zNew = int(ps[1] > .5)
        else:
            #sample z ~ p(z | x, ...)
            zNew = bernoulli(ps[1])

        zs[ii] = zNew
        #recorrect sufficient stats
        ni[zNew] += 1
        pData = 1.0
        #contribution of p(x | z, ...) to ll
        for coord in [0,1]:
            pData *= norm.pdf(xi[coord], ms[zNew][coord],
                              vs[zNew][coord][coord])
        ll += log(pData + 1e-20)

    n1 = sum(zs)
    n0 = n - n1

    #contribution of p(z) to ll (the integral over the posterior Beta)
    zTerm = gammaln(n0 + u) + gammaln(n1 + v) - gammaln(n0 + n1 + u + v)
    return zTerm + ll

def jointLL(xs, ms, vs, zs, sigma, a, b, u, v):
    #compute joint ll of model state
    ll = 0

    #p(mu)
    ll += log(norm.pdf(ms[0][0], 0, sigma) + 1e-20) +\
          log(norm.pdf(ms[0][1], 0, sigma) + 1e-20)
    ll += log(norm.pdf(ms[1][0], 0, sigma) + 1e-20) +\
          log(norm.pdf(ms[1][1], 0, sigma) + 1e-20)

    #p(Sigma)
    #the 1e-100 prevents log(0) when param values are really bad
    ll += log(gammaDist.pdf(1/vs[0][0][0], a, 1/b) + 1e-100)
    ll += log(gammaDist.pdf(1/vs[0][1][1], a, 1/b) + 1e-100)
    ll += log(gammaDist.pdf(1/vs[1][0][0], a, 1/b) + 1e-100)
    ll += log(gammaDist.pdf(1/vs[1][1][1], a, 1/b) + 1e-100)

    #p(x|z, mu, Sigma)
    for zi,xi in zip(zs, xs):
        m = ms[zi]
        var = vs[zi]
        ll += log(norm.pdf(xi[0], m[0], var[0][0]) + 1e-100)
        ll += log(norm.pdf(xi[1], m[1], var[1][1]) + 1e-100)
    n1 = sum(zs)
    n0 = len(zs) - n1
    #p(z|u, v) = integral p(z|pi)p(pi|u,v) dpi
    # = integral Beta(pi ; u + n0, v + n1) dpi
    zTerm = gammaln(n0 + u) + gammaln(n1 + v) - gammaln(n0 + n1 + u + v)
    return ll + zTerm

def sampleMV(xs, ms, vs, zs, sigma, a, b, u, v, sigmaQ=.1):
    accM = 0
    rejM = 0
    accV = 0
    rejV = 0

    for cluster in [0,1]:
        for coord in [0,1]:
            oldL = jointLL(xs, ms, vs, zs, sigma, a, b, u, v)
            oldM = ms[cluster][coord]

            if "guided-mh" in OPTIONS:
                #proposal centered between x_old and the empirical mean

                #calculate cluster mean
                clusterMean = 0
                tot = 0
                for xi,zi in zip(xs, zs):
                    if zi == cluster:
                        clusterMean += xi[coord]
                        tot += 1
                if tot > 0:
                    clusterMean /= tot

                #move about 1 std dev toward the empirical mean
                interp = sigmaQ / abs(clusterMean - oldM)
                proposalMean = (oldM + interp * clusterMean) / (1 + interp)

                newM = normal(proposalMean, sigmaQ)
                #proposal distribution given x_new
                proposalMeanBack = (newM + interp * clusterMean) / (1 + interp)

                qNew = log(norm.pdf(newM, proposalMean, sigmaQ) + 1e-100)
                qOld = log(norm.pdf(oldM, proposalMeanBack, sigmaQ) + 1e-100)
            else:
                #symmetric normal proposal, so q factors cancel
                newM = normal(oldM, sigmaQ)
                #dummy values
                qNew = 0
                qOld = 0

            ms[cluster][coord] = newM
            newL = jointLL(xs, ms, vs, zs, sigma, a, b, u, v)

            try:
                if "reverse-mhr" in OPTIONS:
                    #this block gets the MH ratio wrong
                    mhr = exp(oldL - newL)
                else:
                    mhr = exp(newL - oldL + qOld - qNew)
            except OverflowError:
                mhr = 1

            if "greedy-mh" in OPTIONS:
                #greedy MH
                accept = newL > oldL
            else:
                #normal stochastic MH
                accept = bernoulli(mhr)

            if not accept:
                ms[cluster][coord] = oldM
                rejM += 1
            else:
                accM += 1

            oldL = jointLL(xs, ms, vs, zs, sigma, a, b, u, v)
            oldV = vs[cluster][coord][coord]

            if "guided-mh" in OPTIONS:
                #calculate empirical variance
                clusterVar = 0
                tot = 0
                for xi,zi in zip(xs, zs):
                    if zi == cluster:
                        clusterVar += (xi[coord] - clusterMean)**2
                        tot += 1
                if tot - 1 > 0:
                    clusterVar /= (tot - 1)
                else:
                    clusterVar = 1

                interp = sigmaQ / (abs(clusterVar - oldV) + .01)
                proposalMean = (oldV + interp * clusterVar) / (1 + interp)
                newV = normal(proposalMean, sigmaQ)
                proposalMeanBack = (newV + interp * clusterVar) / (1 + interp)
                qNew = log(norm.pdf(newV, proposalMean, sigmaQ) + 1e-100)
                qOld = log(norm.pdf(oldV, proposalMeanBack, sigmaQ) + 1e-100)
            else:
                newV = normal(oldV, sigmaQ)
                qNew = 0
                qOld = 0

            vs[cluster][coord][coord] = newV
            newL = jointLL(xs, ms, vs, zs, sigma, a, b, u, v)
            #symmetric proposal, so q factors cancel
            try:
                if "reverse-mhr" in OPTIONS:
                    #this block gets the MH ratio wrong
                    mhr = exp(oldL - newL)
                else:
                    mhr = exp(newL - oldL + qOld - qNew)
            except OverflowError:
                mhr = 1

            if "greedy-mh" in OPTIONS:
                #greedy MH
                accept = newL > oldL
            else:
                #normal stochastic MH
                accept = bernoulli(mhr)

            if not accept:
                vs[cluster][coord][coord] = oldV
                rejV += 1
            else:
                accV += 1

    return (accM, rejM, accV, rejV)

if __name__ == "__main__":
    inf = file(sys.argv[1])
    msTrue = pickle.load(inf)
    vsTrue = pickle.load(inf)
    piTrue = pickle.load(inf)
    zsTrue = pickle.load(inf)
    xs = pickle.load(inf)
    ll1True = pickle.load(inf)
    ll2True = pickle.load(inf)

    zs = []
    for pt in xs:
        if "single-cluster" in OPTIONS:
            #bad initializer-- all in one cluster
            zs.append(0)
        else:
            #randomly initialize z
            zs.append(int(random() < .5))

    #initialize clusters to standard normal at origin
    m1 = [0,0]
    m2 = [0,0]
    ms = [m1, m2]
    v1 = [[1,0],[0,1]]
    v2 = [[1,0],[0,1]]
    vs = [v1, v2]

    if "true-params" in OPTIONS:
        #start at true mu, sigma
        ms = msTrue
        vs = vsTrue

    if "far-away" in OPTIONS:
        m1 = [-10, -10]
        m2 = [0, 0]
        ms = [m1, m2]

    if "bad-init" in OPTIONS:
        #find an outlier point and put cluster 0 on top of it
        out = None
        dist = 0
        dmean = [0, 0]
        for pt in xs:
            dmean[0] += pt[0]
            dmean[1] += pt[1]
            d2 = sqrt(pt[0]**2 + pt[1]**2)
            if d2 > dist:
                dist = d2
                out = pt
        ms[0] = out
        vs[0] = [[.001, 0],[0, .001]]
        #set the other cluster to data mean and var
        dmean[0] /= len(xs)
        dmean[1] /= len(xs)
        dvar = [0, 0]
        for pt in xs:
            dvar[0] += (pt[0] - dmean[0])**2
            dvar[1] += (pt[1] - dmean[1])**2
        dvar[0] /= (len(xs) - 1)
        dvar[1] /= (len(xs) - 1)
        ms[1] = dmean
        vs[1] = [[dvar[0], 0], [0, dvar[1]]]
        #print "mean", dmean
        #print "var", dvar

    if "sigmaq-small" in OPTIONS:
        #proposal kernel way too small
        #leads to slow convergence
        sigmaQ = .01
    elif "sigmaq-large" in OPTIONS:
        #proposal kernel too big
        #also slow convergence
        sigmaQ = 10
    else:
        sigmaQ = .1

    lls = []
    accM = 0
    rejM = 0
    accV = 0
    rejV = 0
    sm0 = []
    sm1 = []
    steps = 100

    sigma = 1
    a = 1
    b = 1
    u = 1
    v = 1
    
    for step in range(steps):
        print step
        if "greedy-mh" not in OPTIONS or step > 0:
            sampleZ(xs, ms, vs, zs, u, v)
        if "true-params" not in OPTIONS:
            (aM, rM, aV, rV) = sampleMV(xs, ms, vs, zs,
                                        sigma, a, b, u, v, sigmaQ=sigmaQ)

            accM += aM
            rejM += rM
            accV += aV
            rejV += rV

        sm0.append(copy(ms[0]))
        sm1.append(copy(ms[1]))
        lls.append(jointLL(xs, ms, vs, zs, sigma, a, b, u, v))

    print "Means accepted", accM, "rejected", rejM, "ratio", accM/(accM+rejM)
    print "Vars accepted", accV, "rejected", rejV, "ratio", accV/(accV+rejV)

    print "Final ll", lls[-1], "true ll (integrated)", ll2True

    pylab.plot(range(steps), lls)
    pylab.plot(range(steps), [ll2True for x in range(steps)], 'r')
    sm0 = array(sm0)
    sm1 = array(sm1)
    plotData(ms, vs, zs, xs)
    pylab.plot(sm0[:,0], sm0[:,1], 'r')
    pylab.plot(sm1[:,0], sm1[:,1], 'b')
    plotData(msTrue, vsTrue, zsTrue, xs)
    pylab.show()
